﻿using System.Collections;

namespace Kompozyt
{
    /*public interface IPracownik
    {
        public void WyświetlInformacje(int głębokość);
    }*/

    public interface IOdwiedzający
    {
        public void Odwiedź(Pracownik pracownik, int głębokość);
    }

    public class Pracownik //liść
    {
        private string imię, nazwisko, stanowisko;

        protected internal bool odwiedzony = false;

        public Pracownik(string imię, string nazwisko, string stanowisko)
        {
            this.imię = imię;
            this.nazwisko = nazwisko;
            this.stanowisko = stanowisko;
        }

        public override string ToString()
        {
            return $"{imię} {nazwisko} ({stanowisko})";
        }

        /*
        public virtual void WyświetlInformacje(int głębokość)
        {
            for (int i = 0; i < głębokość; ++i) Console.Write("  ");
            Console.WriteLine($"{imię} {nazwisko} ({stanowisko})");
            odwiedzony = true;
        }
        */

        public virtual void PrzyjmijWizytę(IOdwiedzający odwiedzający, int głębokość)
        {
            odwiedzający.Odwiedź(this, głębokość);
            odwiedzony = true;
        }

        public virtual void Resetuj()
        {
            odwiedzony = false;
        }
    }

    public class Kierownik : Pracownik, IEnumerable<Pracownik> //gałąź, węzeł --> kompozyt (umię poradzić sobie z cyklami)
    {
        private List<Pracownik> podwładni = new List<Pracownik>();

        public Kierownik(string imię, string nazwisko, string stanowisko)
            :base(imię, nazwisko, stanowisko)
        {
        }

        public void DodajPodwładnego(Pracownik pracownik)
        {
            if(!podwładni.Contains(pracownik))
                podwładni.Add(pracownik);
        }

        private static int? głębokośćNajwyższegoKierownika = null;

        /*
        public override void WyświetlInformacje(int głębokość = 0)
        {
            if (głębokośćNajwyższegoKierownika == null)
                głębokośćNajwyższegoKierownika = głębokość;

            base.WyświetlInformacje(głębokość);
            foreach (Pracownik pracownik in podwładni)
                if (!pracownik.odwiedzony)
                    pracownik.WyświetlInformacje(głębokość + 1);
                else Console.WriteLine("Cykl!");

            if (głębokość == głębokośćNajwyższegoKierownika) Resetuj();
        }
        */

        public override void PrzyjmijWizytę(IOdwiedzający odwiedzający, int głębokość = 0)
        {
            if (głębokośćNajwyższegoKierownika == null) głębokośćNajwyższegoKierownika = głębokość;

            base.PrzyjmijWizytę(odwiedzający, głębokość);
            foreach (Pracownik pracownik in podwładni)
                if (!pracownik.odwiedzony)
                    pracownik.PrzyjmijWizytę(odwiedzający, głębokość + 1);                

            if (głębokość == głębokośćNajwyższegoKierownika) Resetuj();
        }

        public override void Resetuj()
        {
            base.Resetuj();
            foreach (Pracownik pracownik in podwładni)                
                if(pracownik.odwiedzony) pracownik.Resetuj();
        }

        public IEnumerator<Pracownik> GetEnumerator()
        {
            /*
            OdwiedzającySpłaszczającyDoListy o = new OdwiedzającySpłaszczającyDoListy();
            PrzyjmijWizytę(o);
            return o.Lista.GetEnumerator();
            */

            return new IteratorPracownik(this);
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }

    public class IteratorPracownik : IEnumerator<Pracownik>
    {
        private uint indeksPracownika = 0;

        Pracownik[] pracownicy = null;

        public IteratorPracownik(Kierownik kierownik)
        {
            OdwiedzającySpłaszczającyDoListy o = new OdwiedzającySpłaszczającyDoListy();
            kierownik.PrzyjmijWizytę(o);
            List<Pracownik> lista = new List<Pracownik>();
            lista.Add(kierownik);
            lista.AddRange(o.Lista);
            pracownicy = lista.ToArray();
        }

        public Pracownik Current => pracownicy[indeksPracownika];

        object IEnumerator.Current => Current;

        public void Dispose()
        {
            //
        }

        public bool MoveNext()
        {
            bool czyOstatni = indeksPracownika == pracownicy.Length - 1;
            if(!czyOstatni) indeksPracownika++;
            return !czyOstatni;
        }

        public void Reset()
        {
            indeksPracownika = 0;
        }
    }

    public class OdwiedzającyWyświetlającyInformacje : IOdwiedzający
    {
        public void Odwiedź(Pracownik pracownik, int głębokość)
        {
            for (int i = 0; i < głębokość; ++i) Console.Write("  ");
            Console.WriteLine(pracownik);
        }
    }

    public class OdwiedzającySpłaszczającyDoListy : IOdwiedzający
    {
        private List<Pracownik> lista = new List<Pracownik>();

        public void Odwiedź(Pracownik pracownik, int głębokość)
        {
            lista.Add(pracownik);
        }

        public List<Pracownik> Lista { get { return lista; } }
    }

    public class OdwiedzającyZliczający : IOdwiedzający
    {
        public uint Licznik { get; private set; } = 0;

        public void Odwiedź(Pracownik pracownik, int głębokość)
        {
            Licznik++;
        }


    }
}
